#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

static int par[4000], npar;
#define maxpar (sizeof par/sizeof*par)

typedef struct T {
  int n;
  struct { int x, y, w, h; } f, t;
  struct { int x, y; } d, a;
} T;

static T trans[1000];
static int ntrans;
#define maxtrans (sizeof trans/sizeof*trans)

#define LEFT par[0]
#define BOTTOM par[1]
#define RIGHT par[2]
#define TOP par[3]

#define START (par + 4)
#define GOAL (par + 6)

static void die (const char * fmt, ...) {
  va_list ap;
  fflush (stdout);
  fputs ("*** check-mcracetrack-solution: ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static void msg (const char * fmt, ...) {
  va_list ap;
  fputs ("[check-mcracetrack-solution] ", stdout);
  va_start (ap, fmt);
  vfprintf (stdout, fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
}

static int inbox (int x, int y) {
  return LEFT < x && x < RIGHT && BOTTOM < y && y < TOP;
}

#define SWAP(TYPE,A,B) \
do { \
  TYPE TMP = (A); \
  (A) = (B);  \
  (B) = TMP; \
} while (0)

static int intersects (int * s, int * t) {
  double d, e, de, x, y;
  if (s[0] > s[2]) { SWAP (int, s[0], s[2]); SWAP (int, s[1], s[3]); }
  if (t[0] > t[2]) { SWAP (int, t[0], t[2]); SWAP (int, t[1], t[3]); }
  if (s[0] == s[2] && t[0] == t[2]) {
    if (s[0] != t[0]) return 0;
    if (t[2] < s[0]) return 0;
    if (s[2] < t[0]) return 0;
    return 1;
  }
  if (s[0] == s[2] && s[1] > s[3]) SWAP (int, s[1], s[3]);
  if (t[0] == t[2] && t[1] > t[3]) SWAP (int, t[1], t[3]);
  if (s[0] == s[2]) SWAP (int *, s, t);
  assert (s[0] < s[2]);
  assert (t[0] <= t[2]);
  d = (s[3] - s[1]) / (double) (s[2] - s[0]);
  if (t[0] == t[2]) {
    if (s[2] < t[0]) return 0;
    if (t[2] < s[0]) return 0;
    y = d * (t[0] - s[0]) + s[1];
    assert (t[1] <= t[3]);
    if (y < (double) t[1]) return 0;
    if (y > (double) t[3]) return 0;
    if (s[1] <= y && y <= s[3]) return 1;
    if (s[3] <= y && y <= s[1]) return 1;
    return 0;
  }
  assert (t[0] < t[2]);
  e = (t[3] - t[1]) / (double) (t[2] - t[0]);
  de = d - e;
  if (!de) {
    if (s[1] - d*s[0] != t[1] - e*t[0]) return 0;
    if (s[2] < t[0]) return 0;
    if (t[2] < s[0]) return 0;
    return 1;
  }
  x = t[1] - s[1] + d*s[0] - e*t[0];
  if (x < (double) s[0]) return 0;
  if ((double) s[2] < x) return 0;
  return 1;
}

static FILE * parcours_file, * solution_file;

int main (int argc, char ** argv) {
  int i, ch;
  if (argc != 3) die ("expected two arguments");
  if (!(parcours_file = fopen (argv[1], "r")))
    die ("can not read parcours file '%s'", argv[1]);
  msg ("reading parcours file '%s'", argv[1]);
  if (!(solution_file = fopen (argv[2], "r")))
    die ("can not read solution file '%s'", argv[2]);
  msg ("reading solution file '%s'", argv[2]);
  for (;;) {
    if (npar == maxpar) die ("too many line segments in '%s'", argv[1]);
    if (fscanf (parcours_file, "%d", par + npar) != 1) break;
    npar++;
  }
  if (npar < 4) die ("bounding box missing in '%s'", argv[1]);
  if (npar < 6) die ("start missing in '%s'", argv[1]);
  if (npar < 8) die ("goal missing in '%s'", argv[1]);
  fclose (parcours_file);
  if ((npar & 3)) die ("%d segments missing", 4 - (npar & 3));
  msg ("bounding box %d %d %d %d", LEFT, BOTTOM, RIGHT, TOP);
  msg ("start-goal %d %d %d %d", START[0], START[1], GOAL[0], GOAL[1]);
  for (i = 8; i < npar; i+= 4)
    msg ("segment[%d] %d %d %d %d",
      (i-8)/4 + 1, par[i], par[i+1], par[i+2], par[i+3]);
  for (;;) {
    ch = getc (solution_file);
    if (ch == 's') break;
    while ((ch = getc (solution_file)) != '\n')
      if (ch == EOF) die ("unexpected end-of-file in '%s'", argv[2]);
  }
  if (getc (solution_file) != ' ') die ("expected space after 's'");
  ch = getc (solution_file);
  if (ch == 'U') {
    if (getc (solution_file) != 'N' ||
        getc (solution_file) != 'R' ||
        getc (solution_file) != 'E' ||
        getc (solution_file) != 'A' ||
        getc (solution_file) != 'C' ||
        getc (solution_file) != 'H' ||
        getc (solution_file) != 'A' ||
        getc (solution_file) != 'B' ||
        getc (solution_file) != 'L' ||
        getc (solution_file) != 'E' ||
        getc (solution_file) != '\n')
INVALID_RESULT_LINE:
      die ("invalid result line in '%s'", argv[2]);
    msg ("found 's UNREACHABLE' result line (skipping rest)");
    fclose (solution_file);
  } else {
    if (ch != 'R' ||
        getc (solution_file) != 'E' ||
        getc (solution_file) != 'A' ||
        getc (solution_file) != 'C' ||
        getc (solution_file) != 'H' ||
        getc (solution_file) != 'E' ||
        getc (solution_file) != 'D' ||
        getc (solution_file) != '\n')
      goto INVALID_RESULT_LINE;
    msg ("found 's REACHABLE' thus checking solution");
    for (;;) {
      T t;
      if (ntrans == maxtrans) die ("too many solution lines");
      if (fscanf (solution_file,
	   "%d: %d %d %d %d --[%d,%d]-> %d %d %d %d (%d,%d)",
	   &t.n,
	   &t.f.x, &t.f.y, &t.f.w, &t.f.h,
	   &t.d.x, &t.d.y,
	   &t.t.x, &t.t.y, &t.t.w, &t.t.h,
	   &t.a.x, &t.a.y) != 13) break;
      trans[ntrans++] = t;
    }
    fclose (solution_file);
    if (!ntrans) {
      if (START[0] != GOAL[0] || START[1] != GOAL[1])
	die ("no solutions found (but start and goal different)");
      msg ("no solution lines since start equal to goal");
    } else {
      msg ("found %d solution lines", ntrans);
      if (trans[0].f.x != START[0])
	die ("start x-coordinate mismatch");
      if (trans[0].f.y != START[1])
	die ("start y-coordinate mismatch");
      msg ("checked first transition begins at start state");
      if (trans[ntrans-1].t.x != GOAL[0])
	die ("goal x-coordinate mismatch");
      if (trans[ntrans-1].t.y != GOAL[1])
	die ("goal y-coordinate mismatch");
      msg ("checked last transition leads to goal state");
      for (i = 0; i < ntrans; i++)
	if (trans[i].n != i + 1)
	  die ("transition %d numbered %d", i + 1, trans[i].n);
      msg ("checked transitions to be numbered correctly");
      for (i = 0; i + 1 < ntrans; i++) {
	T c = trans[i], n = trans[i+1];
	if (c.t.x != n.f.x ||
	    c.t.y != n.f.y ||
	    c.t.w != n.f.w ||
	    c.t.h != n.f.h)
	  die ("transition %d and %d not connected", c.n, n.n);
      }
      msg ("checked transitions to be connected");
      for (i = 0; i < ntrans; i++) {
	T t = trans[i];
	if (t.t.x != t.f.x + t.d.x)
	  die ("incorrect x-difference in transition %d", t.n);
	if (t.t.y != t.f.y + t.d.y)
	  die ("incorrect y-difference in transition %d", t.n);
	if (t.t.w != t.d.x)
	  die ("incorrect x-velocity in to-state of transition %d", t.n);
	if (t.t.h != t.d.y)
	  die ("incorrect y-velocity in to-state of transition %d", t.n);
      }
      msg ("checked local consistency of all transitions");
      for (i = 0; i < ntrans; i++) {
	T t = trans[i];
	if (!inbox (t.f.x, t.f.y))
	  die ("from-state of transition %d out of bounding box", t.n);
	else if (!inbox (t.t.x, t.t.y))
	  die ("to-state of transition %d out of bounding box", t.n);
      }
      msg ("checked all states to be inside of bounding box");
      for (i = 0; i < ntrans; i++) {
	int t[4] = { trans[i].f.x, trans[i].f.y, trans[i].t.x, trans[i].t.y }, j;
	for (j = 8; j < npar; j += 4) {
	  int s[4] = { par[j], par[j+1], par[j+2], par[j+3] };
	  if (intersects (t, s))
	    die ("transition %d and segment %d intersect",
	      trans[i].n, (j-8)/4+1);
	}
      }
      msg ("checked no intersection with segments");
    }
  }
  return 0;
}
