/*------------------------------------------------------------------------*/
/* Copyright (C) 2015 Armin Biere Johannes Kepler University Linz Austria */
/*------------------------------------------------------------------------*/

#ifndef _common_h_INCLUDED
#define _common_h_INCLUDED

#include <stdio.h>

/*------------------------------------------------------------------------*/

#define SWAP(TYPE,A,B) \
do { \
  TYPE TMP = (A); \
  (A) = (B); \
  (B) = TMP; \
} while (0)

/*------------------------------------------------------------------------*/
#ifndef NLOG
void set_logging_enabled ();
int is_logging_enabled ();
#define LOGMSG(FMT,ARGS...) \
do { \
  if (!is_logging_enabled ()) break; \
  msg (FMT, ##ARGS); \
} while (0)
#else
#define LOGMSG(ARGS...) do { } while (0)
#endif
/*------------------------------------------------------------------------*/

extern int start[2], goal[2], box[4];
extern int * segs, * top_segs;

/*------------------------------------------------------------------------*/

extern size_t max_allocated_bytes;

/*------------------------------------------------------------------------*/

void * allocate (size_t);
void * callocate (size_t);
void * reallocate (void *, size_t, size_t);
void deallocate (void *, size_t);

/*------------------------------------------------------------------------*/

const char * read_parcours (FILE *);
void set_default_parcours ();
void reset_parcours ();

/*------------------------------------------------------------------------*/

void set_program_name (const char *);
void set_common_maxaccel (int m);
void set_use_wall_clock_time ();
void set_quiet_mode ();

/*------------------------------------------------------------------------*/

void die (const char * fmt, ...);
void msg (const char * fmt, ...);

/*------------------------------------------------------------------------*/

double percent (double, double);
double average (double, double);
const char * b2s (size_t bytes);
double get_time ();

/*------------------------------------------------------------------------*/

int is_valid_point (int cx, int cy);
int is_valid_move (int cx, int cy, int nx, int ny);
void print_check_valid_move_statistics ();

/*------------------------------------------------------------------------*/

const char * version_string ();
const char * compile_time_options ();

/*------------------------------------------------------------------------*/
#endif
